/*
 * driver_kdvi_x68.c Ver.0.6 for kdvi.
 *
 * Driver module of kdvi.
 *
 * This is a driver that shows bitmap image on console screen.
 * This is for NetBSD/x68k.
 *
 * Copyright 1994,1995 by Takuya Koumoto.
 *
 * Ver.0.0 (FreeBSD(98)) to 0.0 (NetBSD/x68k) :
 * Ver.0.1a(FreeBSD(98)) to 0.1a(NetBSD/x68k) :
 * Ver.0.1a(NetBSD/x68k) to 0.1b(NetBSD/x68k) :
 * Ver.0.2 (FreeBSD(98)) to 0.2 (NetBSD/x68k) :
 * Ver.0.2 (NetBSD/x68k) to 0.2a(NetBSD/x68k) : Modified 1994 by Kunihiko Hayashi.
 * Ver.0.4 (FreeBSD(98)) to 0.4 (NetBSD/x68k) :
 * Ver.0.6 (FreeBSD(98)) to 0.6 (NetBSD/x68k) : Modified 1995 by Kunihiko Hayashi.
 *
 * All rights reserved.
 */

#undef  EXEC_MACHINE
#define EXEC_MACHINE     "x68k"

#define DVI_EXT ".dvi"
#define MAG_FLAG "-mag"
#define DPI_FLAG "-dpi"
#define FROMPAGE_FLAG "-f"
#define FONT_TEST_FLAG "-l"
#define SCROLL_FLAG      "-s"
#define FORECOLOR_FLAG   "-fg"
#define BACKCOLOR_FLAG   "-bg"
#define MAX_PAGE_STORAGE 5

#define DISP_WIDTH    768
#define DISP_HEIGHT   512
#define KDVI_WIDTH    1024
#define KDVI_HEIGHT   1024
#define SCROLL_LIMIT_WIDTH  KDVI_WIDTH -DISP_WIDTH
#define SCROLL_LIMIT_HEIGHT KDVI_HEIGHT-DISP_HEIGHT

#define VRAM_OFFSET0 0x004000      /* X68k VRAM P0 offset */
#define VRAM_OFFSET1 0x024000      /* X68k VRAM P1 offset */
#define VRAM_OFFSET2 0x044000      /* X68k VRAM P2 offset */
#define VRAM_OFFSET3 0x064000      /* X68k VRAM P3 offset */
#define VRAM_LENGTH  0x020000      /* X68k VRAM length of 1 plane */
#define VRAM_TOTAL_LENGTH 0x084000 /* X68k VRAM total length */
#define TEXT_PALLET  0x002200      /* X68k Color pallet port */
#define VIDEO_PORT   0x002600      /* X68k Video-Controller port */
#define CRTC_TEXT_SCROLL  0x000014 /* X68k TEXT-VRAM scroll */

#define KEY_QUIT 'q'
#define KEY_NEXT_PAGE 'n'
#define KEY_PREV_PAGE 'p'
#define KEY_GOTO_PAGE 'g'
#define KEY_UP 'j'
#define KEY_DOWN 'k'
#define KEY_RIGHT 'l'
#define KEY_LEFT 'h'
#define KEY_UP2 'J'
#define KEY_DOWN2 'K'
#define KEY_RIGHT2 'L'
#define KEY_LEFT2 'H'
#define KEY_BACK 'b'
#define KEY_FORW 'f'
#define KEY_REVERSE 'r'
#define KEY_SPACE ' '
#define KEY_HELP '?'
#define KEY_HELP2 '/'
#define KEY_INFO 'i'
#define KEY_SUSPEND 0x1a /* ^Z */
#define KEY_STOP 0x03    /* ^C */
#define KEY_EOF 0x04     /* ^D */
#define KEY_REDRAW 'R'
#define KEY_ESC 0x1B

#include "config.h"
#include "typedef.h"
#include "enter.h"
#include "vars.h"
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/file.h>
#ifdef HAVE_UNISTD
#include <unistd.h>
#endif
#include <sys/ioctl.h>
#include <sgtty.h>
#include <signal.h>
#include "version.h"
#include "bitmap.h"
#include "buffer.h"
#include "file.h"
#include "fontlist.h"
#include "init.h"
#include "dvi.h"

/* Local variables */
static int count = 0;
static long init_flag = 0;
static int vram_fd = 0;
static u_char *vram = NULL;
static u_char vram_name[] = "/dev/grf0";  /* NetBSD/x68k VRAM device */
static u_short old_pallet[16];

static long color_white_back = 0;
static long kdvi_width = KDVI_WIDTH;
static long kdvi_height = KDVI_HEIGHT;
static long kdvi_width_8 = 0;
static long x_offset_8 = 0;
static long y_offset = 0;
static long x_step_8 = 0;
static long y_step = 0;
static BUFFER *buffer_array[MAX_PAGE_STORAGE];
static long last_buffer = 0;
static long dvi_mtime = 0;
#define COLOR_BLACK  0x0000
#define COLOR_WHITE  0xffff
static long disp_width = KDVI_WIDTH;
static long disp_width_8 = 0;
static long scroll_step = 16;
static long scroll_flg = 0;
static long fore_color = COLOR_WHITE;
static long back_color = COLOR_BLACK;
static struct { short x,y; } textpos;

/* Prototype of global functions */
void Init_driver(int, char**);
void Show_page(void);
long Reset_dimension(void);

/* Prototype of local functions */
static void set_name(int, char**);
static void set_mag(int, char**);
static void set_dpi(int, char**);
static void set_begin_page(int, char**);
static void set_font_test(int, char**);
static void set_scroll_mode(int, char**);
static void set_fore_color(int, char**);
static void set_back_color(int, char**);

static void usage(void);
static void copyright(void);
static void show_help(void);
static void show_information(void);

static void gr_open(void);
static void gr_screen_on(void);
static void gr_screen_off(void);
static void gr_cls(void);
static void gr_draw(void);
static void gr_end(void);
static void gr_init_color(void);
static void gr_reset_color(void);
static void gr_reverse_color(void);
static void gr_set_pallet(long,u_short);
static u_short gr_get_pallet(long);

static void suspend(void);
static int getch(void);

static void gr_init_scroll(void);
static int gr_move_scroll(short,short);

/* Global functions */
void Init_driver(int argc, char **argv)
{
  static long file_name_flag = 0;

  Enter("Init_driver");

  if (argc < 2)
    {
      copyright();
      usage();
      exit(1);
    }
  count=1;
  while(count < argc)
    {
      if (argv[count][0] == '-')
	{
	  if ((strcmp(argv[count], MAG_FLAG)) == 0)
	    {
	      set_mag(argc, argv);
	    }
	  else if ((strcmp(argv[count], DPI_FLAG)) == 0)
	    {
	      set_dpi(argc, argv);
	    }
	  else if ((strcmp(argv[count], FROMPAGE_FLAG)) == 0)
	    {
	      set_begin_page(argc, argv);
	    }
	  else if ((strcmp(argv[count], FONT_TEST_FLAG)) == 0)
	    {
	      set_font_test(argc, argv);
	    }
	  else if ((strcmp(argv[count], SCROLL_FLAG)) == 0) /***new***/
	    {
	      set_scroll_mode(argc, argv);
	    }
	  else if ((strcmp(argv[count], FORECOLOR_FLAG)) == 0) /***new***/
	    {
	      set_fore_color(argc, argv);
	    }
	  else if ((strcmp(argv[count], BACKCOLOR_FLAG)) == 0) /***new***/
	    {
	      set_back_color(argc, argv);
	    }
	  else
	    {
	      copyright();
	      usage();
	      exit(1);
	    }
	}
      else
	{
	  set_name(argc, argv);
	  file_name_flag++;
	}
    }
  if (file_name_flag == 0)
    {
      copyright();
      fprintf(stderr, "Init_driver: There is no dvifiles.\n");
      usage();
      exit(1);
    }
  if (scroll_flg == 0) {
    disp_width = DISP_WIDTH;
    kdvi_width = KDVI_WIDTH;
    kdvi_height = DISP_HEIGHT;
  }
  else {
    disp_width = KDVI_WIDTH;
    kdvi_width = KDVI_WIDTH;
    kdvi_height = KDVI_HEIGHT;
  }
  strcpy(Sys_env.pk_path, PKDIR);
  strcpy(Sys_env.tfm_path, TFMDIR);
  Dim.h_sp_offset = (long)(H_OFFSET * 72.27 * 65536);
  Dim.v_sp_offset = (long)(V_OFFSET * 72.27 * 65536);
  kdvi_width_8 = (kdvi_width / 8);
  disp_width_8 = ((disp_width + 7) / 8);
  x_step_8 = kdvi_width_8 / 4;
  y_step = kdvi_height / 4;
}

void Show_page(void)
{
  Enter("Show_page");

  Draw_box_a_dot(0,0,Buf->width - 1, Buf->height - 1);
  Draw_box_a_dot(1,1,Buf->width - 2, Buf->height - 2);

  if (last_buffer < MAX_PAGE_STORAGE)
    {
      buffer_array[last_buffer] = Buf;
      last_buffer++;
    }

  if (init_flag == 0)
    gr_open();
  gr_draw();
  dvi_mtime = Fmtime(Dvi.full_name);
}

long Reset_dimension(void)
{
  long flag;
  long next_page;
  long page_offset;
  long command;
  long buffer;
  BUFFER * tmp_buf;

  Enter("Reset_dimension");

  flag = 0;
  next_page = Dim.page;
  page_offset = 0;
  
  while (flag == 0)
    {
      command = getch();
/* Cursor Key -> HJKL Key */
      if (command == KEY_ESC)
	{
	  if (getch() == '[') {
	    switch(getch()) {
	      case 'A': command = KEY_DOWN;  break; /* up    arrow */
	      case 'B': command = KEY_UP;    break; /* down  arrow */
	      case 'C': command = KEY_RIGHT; break; /* right arrow */
	      case 'D':	command = KEY_LEFT;  break; /* left  arrow */
	    }
	    fprintf(stderr, "\x1b[%d;%dH[%ld]",
                   (textpos.y/16)+1, (textpos.x/8)+1, Dim.page);
	  }
	}
      if (command == KEY_QUIT)
	{
	  flag++;
	  next_page = -1;
	}
      else if ((command == KEY_NEXT_PAGE) ||
	       (command == KEY_FORW) ||
	       (command == KEY_SPACE))
	{
	  if (page_offset <= 0)
	    next_page++;
	  else
	    next_page += page_offset;
	  page_offset = 0;
	  if (next_page > Dvi.total_page)
	    next_page = Dvi.total_page;
	  if (next_page != Dim.page)
	    flag++;
	}
      else if ((command == KEY_PREV_PAGE) ||
	       (command == KEY_BACK))
	{
	  if (page_offset <= 0)
	    next_page--;
	  else
	    next_page -= page_offset;
	  page_offset = 0;
	  if (next_page < 1)
	    next_page = 1;
	  if (next_page != Dim.page)
	    flag++;
	}
      else if (command == KEY_GOTO_PAGE)
	{
	  if (page_offset <= 0)
	    next_page = 1;
	  else if (page_offset > Dvi.total_page)
	    next_page = Dvi.total_page;
	  else
	    next_page = page_offset;
	  page_offset = 0;
	  if (next_page != Dim.page)
	    flag++;
	}
      else if (command == KEY_UP)
	{
	  if (scroll_flg == 0) {
	    y_offset += y_step;
	    gr_draw();
	  }
	  else if (gr_move_scroll(0,+scroll_step) != 0) {
	    y_offset += scroll_step;
	    gr_draw();
	  }
	}
      else if (command == KEY_DOWN)
	{
	  if (scroll_flg == 0) {
	    y_offset -= y_step;
	    gr_draw();
	  }
	  else if (gr_move_scroll(0,-scroll_step) != 0) {
	    y_offset -= scroll_step;
	    gr_draw();
	  }
	}
      else if (command == KEY_RIGHT)
	{
	  if (scroll_flg == 0) {
	    x_offset_8 += x_step_8;
	    gr_draw();
	  }
	  else if (gr_move_scroll(+scroll_step,0) != 0) {
	    x_offset_8 += scroll_step;
	    gr_draw();
	  }
	}
      else if (command == KEY_LEFT)
	{
	  if (scroll_flg == 0) {
	    x_offset_8 -= x_step_8;
	    gr_draw();
	  }
	  else if (gr_move_scroll(-scroll_step,0) != 0) {
	    x_offset_8 -= scroll_step;
	    gr_draw();
	  }
	}
      else if (command == KEY_UP2)
	{
	  if (scroll_flg == 0) {
	    y_offset += (y_step << 1);
	    gr_draw();
	  }
	  else if (gr_move_scroll(0,+scroll_step << 1) != 0) {
	    y_offset += (scroll_step << 1);
	    gr_draw();
	  }
	}
      else if (command == KEY_DOWN2)
	{
	  if (scroll_flg == 0) {
	    y_offset -= (y_step << 1);
	    gr_draw();
	  }
	  else if (gr_move_scroll(0,-scroll_step << 1) != 0) {
	    y_offset -= (scroll_step << 1);
	    gr_draw();
	  }
	}
      else if (command == KEY_RIGHT2)
	{
	  if (scroll_flg == 0) {
	    x_offset_8 += (x_step_8 << 1);
	    gr_draw();
	  }
	  else if (gr_move_scroll(+scroll_step << 1,0) != 0) {
	    x_offset_8 += (scroll_step << 1);
	    gr_draw();
	  }
	}
      else if (command == KEY_LEFT2)
	{
	  if (scroll_flg == 0) {
	    x_offset_8 -= (x_step_8 << 1);
	    gr_draw();
	  }
	  else if (gr_move_scroll(-scroll_step << 1,0) != 0) {
	    x_offset_8 -= (scroll_step << 1);
	    gr_draw();
	  }
	}
      else if (command == KEY_REVERSE)
	{
	  gr_reverse_color();
	}
      else if ((command == KEY_STOP) ||
	       (command == KEY_EOF))
	{
	  flag++;
	  next_page = -1;
	}
      else if (command == KEY_SUSPEND)
	{
	  suspend();
	}
      else if ((command == KEY_HELP) ||
	       (command == KEY_HELP2))
	{
	  show_help();
	}
      else if (command == KEY_INFO)
	{
	  show_information();
	}
      else if ((command >= '0') && (command <= '9'))
	{
	  page_offset = (page_offset * 10 + (long)(command - '0'));
	  page_offset = page_offset % 10000;
	}
      else if (command == KEY_REDRAW)
	{
	  fprintf(stderr, "\nRedrawing.");
	  return 0;
	}

      if ((flag != 0) && (next_page > 0))
	{
	  buffer = 0;
	  while ((buffer < last_buffer) && (flag != 0))
	    {
	      if (buffer_array[buffer]->page == next_page)
		{
		  Buf = buffer_array[buffer];
		  x_offset_8 = 0;
		  y_offset = 0;
		  Dim.page = next_page;

		  fprintf(stderr, "->[%ld]\n", Dim.page);
		  gr_draw();

		  flag = 0;
		  page_offset = 0;
		}
	      buffer++;
	    }

	  if (flag == 0)
	    {
	      tmp_buf = buffer_array[buffer - 1];
	      while (buffer < last_buffer)
		{
		  buffer_array[buffer - 1] = buffer_array[buffer];
		  buffer++;
		}
	      buffer_array[buffer - 1] = tmp_buf;
	    }
	}
      if (dvi_mtime < Fmtime(Dvi.full_name))
	{
	  flag++;
	}
    }
  if (next_page == -1)
    {
      gr_end();
      return 1;
    }
  else
    {
      x_offset_8 = 0;
      y_offset = 0;
      Dim.page = next_page;
      fprintf(stderr, "->[%ld]\n", Dim.page);
      if (last_buffer != MAX_PAGE_STORAGE)
	{
	  Buf = NULL;
	  Make_buffer();
	}
      else
	{
	  Buf = buffer_array[0];
	  Buf->page = next_page;
	  for (buffer=0; buffer<(last_buffer - 1) ; buffer++)
	    {
	      buffer_array[buffer] = buffer_array[buffer + 1];
	    }
	  buffer_array[last_buffer - 1] = Buf;
	}

      if (dvi_mtime < Fmtime(Dvi.full_name))
	{
	  Clear_fontlist();
	  Clear_dvi();
	  while (last_buffer > 0)
	    {
	      Buf = buffer_array[last_buffer - 1];
	      Clear_buffer();
	      last_buffer--;
	    }
	  last_buffer = 0;
	  Init_dvi();
	  Set_dimension();
	  Make_fontlist();
	}
      return 0;
    }
}

/* Local functions */
static void set_name(int argc, char **argv)
{
  static long flag = 0;
  static long last_slash = -1;
  long length;
  long position;

  Enter("set_name");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_name: Too many dvifiles.\n");
      usage();
      exit(1);
    }
  flag++;
  
  length = strlen(argv[count]);
  strcpy(Dvi.full_name, argv[count]);
  if ((strcmp((Dvi.full_name + length-4), DVI_EXT)) != 0)
    strcpy((Dvi.full_name + length), DVI_EXT);

  length = strlen(Dvi.full_name);
  for (position=0 ; position<(length-4) ; position++)
    if (Dvi.full_name[position] == '/')
      last_slash = position + 1;

  if (last_slash > 0)
    {
      strncpy(Dvi.path_name, Dvi.full_name, last_slash);
      strcpy(Dvi.file_name, (Dvi.full_name + last_slash));
    }
  else
    strcpy(Dvi.file_name, Dvi.full_name);

  count++;

#ifdef DEBUG
  fprintf(stderr, "full_name : %s\n", Dvi.full_name);
  fprintf(stderr, "file_name : %s\n", Dvi.file_name);
  fprintf(stderr, "path_name : %s\n", Dvi.path_name);
#endif
}

static void set_dpi(int argc, char **argv)
{
  static long flag = 0;

  Enter("set_dpi");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_dpi: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  if (count >= argc)
    {
      copyright();
      fprintf(stderr, "set_dpi: No dpi value.\n");
      usage();
      exit(1);
    }
  Dim.dpi = atol(argv[count]);
  if (Dim.dpi < MINIMUM_DPI)
    {
      copyright();
      fprintf(stderr, "set_dpi: Bad dpi number."
	      " You must set it larger then %d.\n", MINIMUM_DPI-1);
      usage();
      exit(1);
    }
  count++;
#ifdef DEBUG
  fprintf(stderr, "Dim.dpi : %d\n", Dim.dpi);
#endif
}

static void set_begin_page(int argc, char **argv)
{
  static long flag = 0;

  Enter("set_begin_page");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_begin_page: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  if (count >= argc)
    {
      copyright();
      fprintf(stderr, "set_begin_page: No page number.\n");
      usage();
      exit(1);
    }
  Dim.page = atol(argv[count]);
  if (Dim.page <= 0)
    {
      copyright();
      fprintf(stderr, "set_begin_page: Bad page number.\n");
      usage();
      exit(1);
    }
  count++;
}

static void set_mag(int argc, char **argv)
{
  static long flag = 0;

  Enter("set_mag");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_mag: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  if (count >= argc)
    {
      copyright();
      fprintf(stderr, "set_mag: No mag value.\n");
      usage();
      exit(1);
    }
  Dim.mag = atol(argv[count]);
  if (Dim.mag < 0)
    {
      copyright();
      fprintf(stderr, "set_mag: Bad magstep number.\n");
      usage();
      exit(1);
    }
  if (Dim.mag == 0)
    Dim.mag = 1000;
  else if (Dim.mag == 1)
    Dim.mag = 1200;
  else if (Dim.mag == 2)
    Dim.mag = 1440;
  else if (Dim.mag == 3)
    Dim.mag = 1728;
  else if (Dim.mag == 4)
    Dim.mag = 2074;
  else if (Dim.mag == 5)
    Dim.mag = 2488;
  count++;
#ifdef DEBUG
  fprintf(stderr, "Dim.mag : %d\n", Dim.mag);
#endif
}

static void set_font_test(int argc, char **argv)
{
  static long flag = 0;

  Enter("set_font_test");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_font_test: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  Dim.font_test = 1;
#ifdef DEBUG
  fprintf(stderr, "Dim.font_test : %d\n", Dim.font_test);
#endif
}

static void set_scroll_mode(int argc, char **argv)
{
  static long flag = 0;

  Enter("set_scroll_mode");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_scroll_mode: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  scroll_flg++;
#ifdef DEBUG
  fprintf(stderr, "scroll_flg : %d\n", scroll_flg);
#endif
}

static void set_fore_color(int argc, char **argv)
{
  static long flag = 0;

  Enter("set_fore_color");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_fore_color: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  if (count >= argc)
    {
      copyright();
      fprintf(stderr, "set_fore_color: No color number.\n");
      usage();
      exit(1);
    }
  sscanf(argv[count],"%x",&fore_color);
  if ((fore_color < 0)||(fore_color > 0xffff))
    {
      copyright();
      fprintf(stderr, "set_fore_color: Bad color number.\n");
      usage();
      exit(1);
    }
  count++;
}

static void set_back_color(int argc, char **argv)
{
  static long flag = 0;

  Enter("set_back_color");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_back_color: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  if (count >= argc)
    {
      copyright();
      fprintf(stderr, "set_back_color: No color number.\n");
      usage();
      exit(1);
    }
  sscanf(argv[count],"%x",&back_color);
  if ((back_color < 0)||(back_color > 0xffff))
    {
      copyright();
      fprintf(stderr, "set_back_color: Bad color number.\n");
      usage();
      exit(1);
    }
  count++;
}

static void usage(void)
{
  Enter("usage");

  fprintf(stderr, "Usage: kdvi [%s dpi] [%s mag] [%s page] "
	"[%s] [%s] [%s fore-color] [%s back-color] <dvi>\n\n",
          DPI_FLAG, MAG_FLAG, FROMPAGE_FLAG, FONT_TEST_FLAG,
	  SCROLL_FLAG, FORECOLOR_FLAG, BACKCOLOR_FLAG);
}

static void copyright(void)
{
  Enter("copyright");

  fprintf(stderr, "kdvi Ver.%s.%s.(%s)\n\n",
                  VERSION_KDVI_98, VERSION_DVILIB,EXEC_MACHINE);
  fprintf(stderr, "Copyright 1994,1995 by Takuya Koumoto. "
                  "All rights reserved.\n");
  fprintf(stderr, "Modified  1994,1995 by Kunihiko Hayashi.\n\n");

}

static void show_help(void)
{
  Enter("show_help");

  gr_end();
  copyright();
  fprintf(stderr, "Key assinments.\n");
  fprintf(stderr, "  h : Scroll left.       H : High speed scroll left.\n");
  fprintf(stderr, "  j : Scroll down.       J : High speed scroll down.\n");
  fprintf(stderr, "  k : Scroll up.         K : High speed scroll up.\n");
  fprintf(stderr, "  l : Scroll right.      L : High speed scroll right.\n");
  fprintf(stderr, "  r : Reverse color.     g : Go to page.\n");
  fprintf(stderr, "  n, f : Forward page.   p, b : Backward page.\n");
  fprintf(stderr, "  R : Redraw page.\n");
  fprintf(stderr, "  C-z : Suspend program. q, C-c, C-d : Quit program.\n");
  fprintf(stderr, "  Numbers : Set number for n,f,p,b,g.\n");
  fprintf(stderr, "  /, ? : Help.           i : Information about DVI file.\n\n");
  fprintf(stderr, "Current page [%ld].\n\n", Dim.page);
  fprintf(stderr, "Hit any key to return.\n");
  while(getch() == NULL)
    ;
  gr_open();
  gr_draw();
}

static void show_information(void)
{
  Enter("show_information");

  gr_end();
  copyright();
  fprintf(stderr, "Information about DVI file.\n");
  fprintf(stderr, "  File name    : %s.\n", Dvi.file_name);
  fprintf(stderr, "  Log          :%s.\n", Dvi.pre);
  fprintf(stderr, "  Total page   : %ld.\n", Dvi.total_page);
  fprintf(stderr, "  Page size    : %ldx%ld pt.\n\n",
	  (Dvi.u >> 16), (Dvi.l >> 16));
  fprintf(stderr, "Information about buffer.\n");
  fprintf(stderr, "  DPI          : %ld.\n", Dim.dpi);
  fprintf(stderr, "  Buffer size  : %ldx%ld pt (%ldx%ld dot).\n",
	  ((Dvi.u >> 16) + (long)(144 * H_OFFSET)),
	  ((Dvi.l >> 16) + (long)(144 * V_OFFSET)),
	  Buf->width, Buf->height);
  fprintf(stderr, "  Page storage : %ld.\n\n", (long)MAX_PAGE_STORAGE);
  fprintf(stderr, "Hit any key to return.\n");
  while(getch() == NULL)
    ;
  gr_open();
  gr_draw();
}

static void gr_open(void)
{
  int prot;
  int flags;
  
  Enter("gr_open");

  if ((vram_fd = open(vram_name, O_RDWR, 0)) < 0)
    {
      fprintf(stderr, "Can't open %s\n", vram_name);
      exit(1);
    }
  
  prot = PROT_READ | PROT_WRITE;
  flags = MAP_FILE | MAP_SHARED;

  vram = (u_char *)mmap((caddr_t)0, VRAM_TOTAL_LENGTH, prot,
			flags, vram_fd, 0);
  if (vram == (u_char *)(-1))
    {
      fprintf(stderr, "vram mmap fail.\n");
      exit(1);
    }
  
  gr_screen_off();
  gr_init_color();
  gr_cls();
  gr_screen_on();

  init_flag++;
}

static void gr_screen_on(void)
{
  Enter("gr_screen_on");

  *(u_short *)(vram + VIDEO_PORT) |= 0x20;
  /* PC9801 Graphic dispray on */
}

static void gr_screen_off(void)
{
  Enter("gr_screen_off");

  *(u_short *)(vram + VIDEO_PORT) &= ~0x20;
  /* PC9801 Graphic dispray off */
}

static void gr_cls(void)
{
  static long flag = 0;

  Enter("gr_cls");

  fprintf(stderr, "\x1b[1;1H\x1b[J");

  if (vram == NULL)
    return;

  bzero((vram + VRAM_OFFSET0), VRAM_LENGTH);

  bzero((vram + VRAM_OFFSET1), VRAM_LENGTH);
  bzero((vram + VRAM_OFFSET2), VRAM_LENGTH);
  bzero((vram + VRAM_OFFSET3), VRAM_LENGTH);
  flag++;
}

static void gr_draw(void)
{
  long y;
  u_char *tmp_vram;
  u_char *tmp_bitmap;

  Enter("gr_draw");

  if (vram == NULL)
    return;

#if 0
  gr_screen_off();
  gr_cls();
#endif

  if (disp_width_8 >= Buf->width_8)
    {
      if (kdvi_height >= Buf->height)
	{
	  x_offset_8 = 0;
	  y_offset = 0;
	  tmp_vram = vram + VRAM_OFFSET0 + ((disp_width_8 - Buf->width_8) / 2)
	    + (((kdvi_height - Buf->height) / 2) * kdvi_width_8);
	  tmp_bitmap = Buf->bitmap;
	  for (y=0 ; y<Buf->height ; y++)
	    {
	      bcopy(tmp_bitmap, tmp_vram, Buf->width_8);
	      tmp_vram += kdvi_width_8;
	      tmp_bitmap += Buf->width_8;
	    }
	}
      else
	{
	  x_offset_8 = 0;
	  if (y_offset < 0)
	    y_offset = 0;
	  else if ((y_offset + kdvi_height) > Buf->height)
	    y_offset = Buf->height - kdvi_height;
	  tmp_vram = vram + VRAM_OFFSET0 + ((disp_width_8 - Buf->width_8) / 2);
	  tmp_bitmap = Buf->bitmap + (Buf->width_8 * y_offset);
	  for (y=0 ; y<kdvi_height ; y++)
	    {
	      bcopy(tmp_bitmap, tmp_vram, Buf->width_8);
	      tmp_vram += kdvi_width_8;
	      tmp_bitmap += Buf->width_8;
	    }
	}
    }
  else
    {
      if (kdvi_height >= Buf->height)
	{
	  if (x_offset_8 < 0)
	    x_offset_8 = 0;
	  else if ((x_offset_8 + disp_width_8) > Buf->width_8)
	    x_offset_8 = Buf->width_8 - disp_width_8;
	  y_offset = 0;
	  tmp_vram = vram + VRAM_OFFSET0 +
	    ((kdvi_height - Buf->height) / 2) * kdvi_width_8;
	  tmp_bitmap = Buf->bitmap + x_offset_8;
	  for (y=0 ; y<Buf->height ; y++)
	    {
	      bcopy(tmp_bitmap, tmp_vram, disp_width_8);
	      tmp_vram += kdvi_width_8;
	      tmp_bitmap += Buf->width_8;
	    }
	}
      else
	{
	  if (x_offset_8 < 0)
	    x_offset_8 = 0;
	  else if ((x_offset_8 + disp_width_8) > Buf->width_8)
	    x_offset_8 = Buf->width_8 - disp_width_8;
	  if (y_offset < 0)
	    y_offset = 0;
	  else if ((y_offset + kdvi_height) > Buf->height)
	    y_offset = Buf->height - kdvi_height;
	  tmp_vram = vram + VRAM_OFFSET0;
	  tmp_bitmap = Buf->bitmap + Buf->width_8 * y_offset + x_offset_8;
	  for (y=0 ; y<kdvi_height ; y++)
	    {
	      bcopy(tmp_bitmap, tmp_vram, disp_width_8);
	      tmp_vram += kdvi_width_8;
	      tmp_bitmap += Buf->width_8;
	    }
	}
    }

#if 0
  gr_screen_on();
#endif
  fprintf(stderr, "\x1b[1;1H\x1b[J[%ld]", Dim.page);
  printf("[%ld]", Dim.page);
}

static void gr_end(void)
{
  Enter("gr_end");

  if (init_flag == 0)
    gr_open();
  gr_screen_off();
  gr_cls();
  gr_reset_color();
  gr_init_scroll();
  munmap(vram, VRAM_TOTAL_LENGTH);
  close(vram_fd);
  init_flag = 0;
}

static void gr_init_color(void)
{
  int i;

  Enter("gr_init_color");

  gr_screen_off();

  /* 16 color mode */

  for(i=0;i<16;i++) {
    old_pallet[i] = gr_get_pallet(i);
  }

  /* 16/4096 color pallete initialize */
  if (color_white_back == 0)
    {
      /* 0 : Black */
      gr_set_pallet(0,back_color);
      /* 1 : Dark blue -> White */
      gr_set_pallet(1,fore_color);
    }
  else
    {
      /* 0 : Black -> White */
      gr_set_pallet(0,fore_color);

      /* 1 : Dark blue -> Black */
      gr_set_pallet(1,back_color);
    }

  gr_screen_on();
}

static void gr_reset_color(void)
{
  int i;
  Enter("gr_reset_color");

  gr_screen_off();

  /* 16/4096 color pallete reset */
  /* 0 : Black */
  /* 1 : Dark blue */

  for(i=0;i<16;i++)
      gr_set_pallet(i,old_pallet[i]);

  gr_screen_on();
}

static void gr_reverse_color(void)
{
  Enter("gr_reverse_color");

  gr_screen_off();
  if (color_white_back == 1)
    {
      /* 0 : White -> Black */
      gr_set_pallet(0,back_color);

      /* 1 : Black -> White */
      gr_set_pallet(1,fore_color);
      
      color_white_back=0;
    }
  else
    {
      /* 0 : Black -> White */
      gr_set_pallet(0,fore_color);
      
      /* 1 : White -> Black */
      gr_set_pallet(1,back_color);
      
      color_white_back = 1;
    }
  gr_screen_on();
}


static void gr_set_pallet(long pal, u_short code)
{
  u_char* pallet_port;

  Enter("gr_set_pallet");

  pallet_port = vram + TEXT_PALLET + ((pal % 16)<<1);
  *(u_short *)pallet_port = code;
}

static u_short gr_get_pallet(long pal)
{
  u_char* pallet_port;

  Enter("gr_get_pallet");

  pallet_port = vram + TEXT_PALLET + ((pal % 16)<<1);
  return *(u_short *)pallet_port;
}


/* This will not work on Version 7. */
static void suspend(void)
{
  void (*oldhandle)();

  Enter("suspend");

  gr_end();
  oldhandle = signal(SIGTSTP, SIG_DFL);
  kill(0, SIGTSTP);
  signal(SIGTSTP, oldhandle);
  gr_open();
  gr_move_scroll(0,0);
  gr_draw();
}

/* This will not work on System V. */
static int getch(void)
{
  static struct sgttyb sgt_old;
  static struct sgttyb sgt_new;
  int key_in_value;

  Enter("getch");
  
  ioctl(0, TIOCGETP, &sgt_old);

  sgt_new = sgt_old;
  sgt_new.sg_flags &= ~ECHO;
  sgt_new.sg_flags |= RAW;

  ioctl(0, TIOCSETP, &sgt_new);

  key_in_value = fgetc(stdin);

  ioctl(0, TIOCSETP, &sgt_old);

  return key_in_value;
}

static void gr_init_scroll()
{
  u_char* textscroll_port;

  Enter("gr_init_scroll");

  textscroll_port = vram + CRTC_TEXT_SCROLL;
  textpos.x = textpos.y = 0;
  *(u_long *)textscroll_port = *(u_long *)&textpos;
}
static int gr_move_scroll(short off_x,short off_y)
{
  int limit_flag = 0;
  u_char* textscroll_port;

  Enter("gr_move_scroll");

  textscroll_port = vram + CRTC_TEXT_SCROLL;

  textpos.x += off_x;
  textpos.y += off_y;

  if (textpos.x < 0) {
    textpos.x = 0;
    limit_flag++;
  }
  else if (textpos.x > SCROLL_LIMIT_WIDTH) {
    textpos.x = SCROLL_LIMIT_WIDTH;
    limit_flag++;
  }
  if (textpos.y < 0) {
    textpos.y = 0;
    limit_flag++;
  }
  else if (textpos.y > SCROLL_LIMIT_HEIGHT) {
    textpos.y = SCROLL_LIMIT_HEIGHT;
    limit_flag++;
  }
  *(u_long *)textscroll_port = *(u_long *)&textpos;

  return limit_flag;
}

/* End of driver_kdvi_x68.c */
