/*
 * font_pk.c Ver.0.7 for kdvi.
 *
 * Handling PK font module of kdvi.
 *
 * Copyright 1994,1995 by Takuya Koumoto.
 *
 * Ver.0.0 to 0.1 : Modified 1994 by Takuya Koumoto.
 * Ver.0.1 to 0.2 : Modified 1994 by Takuya Koumoto.
 * Ver.0.2 to 0.3 : Modified 1995 by Takuya Koumoto.
 * Ver.0.3 to 0.4 : Modified 1995 by Takuya Koumoto.
 * Ver.0.4 to 0.5 : Modified 1995 by Takuya Koumoto.
 * Ver.0.5 to 0.6 : Modified 1995 by Takuya Koumoto.
 * Ver.0.6 to 0.7 : Modified 1995 by Takuya Koumoto.
 *
 * All rights reserved.
 */

/* Maximum character number in one PK font file. */
#define PK_MAX_CHAR_NUMBER 256

#include "config.h"
#include "typedef.h"
#include "enter.h"
#include "vars.h"
#include <sys/file.h>
#ifdef HAVE_UNISTD
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/uio.h>
#include <stdlib.h>
#include "pkcode.h"
#include "conv.h"
#include "bitmap.h"

/* Local variables */
static FONT_INFO *font;
static long c_id;
static int pk_fd;
static double ratio;
static long nyb_count = 0;

/* Prototype of global function */
void Font_pk_decode(void);
void Font_pk_clear(FONT_INFO*);

/* Prototype of local functions */
static void setup_pk_info(void);
static void pk_put_char(void);
static void def_char(long);
static void unpack_raster(void);
static void unpack_run(void);
static long get_nyb(void);

/* Global function */
void Font_pk_decode(void)
{
  Enter("Font_pk_decode");

  font = Code.font;
  c_id = Code.char_id;

  if (font->pk == NULL)
    setup_pk_info();

  Code.sp_width = font->pk->code[c_id].tfm_width;
  pk_put_char();
}

void Font_pk_clear(FONT_INFO *font_info)
{
  long char_code;

  Enter("Font_pk_clear");

  if (font_info->pk != NULL)
    {
      if (font_info->pk->code != NULL)
	{
	  for (char_code=0 ; char_code<PK_MAX_CHAR_NUMBER ; char_code++)
	    {
	      if (font_info->pk->code[char_code].data != NULL)
		free(font_info->pk->code[char_code].data);
	      if (font_info->pk->code[char_code].bitmap != NULL)
		free(font_info->pk->code[char_code].bitmap);
	    }
	  free(font_info->pk->code);
	}
      free(font_info->pk);
    }
}

/* Local functions */
static void setup_pk_info(void)
{
  long code;
  long ds;

  Enter("setup_pk_info");

  pk_fd = open(font->name, O_RDONLY, 0);

  lseek(pk_fd, 0, SEEK_SET);
  if ((Get_u_char(pk_fd)) != PK_PRE)
    {
      fprintf(stderr, "Bad PK file.\n");
      exit(1);
    }
  if ((Get_u_char(pk_fd)) != PK_ID)
    {
      fprintf(stderr, "Bad PK ID.\n");
      exit(1);
    }
  if ((font->pk = calloc(1, sizeof(PK_INFO))) == NULL)
    {
      fprintf(stderr, "setup_pk_info : Can't allocate memory.\n");
      exit(1);
    }
  if ((font->pk->code = calloc(PK_MAX_CHAR_NUMBER, sizeof(PK_CHAR_INFO))) == NULL)
    {
      fprintf(stderr, "setup_pk_info : Can't allocate memory.\n");
      exit(1);
    }
  lseek(pk_fd, ((long)Get_u_char(pk_fd)), SEEK_CUR);
  ds = Get_long(pk_fd);
#ifdef DEBUG
  fprintf(stderr, "ds %ld.\n", (ds >> 20));
#endif
  ratio = ((double)font->attrib->s * (double)ds) /
          ((double)font->attrib->d * (double)(1 << 24));
  
  lseek(pk_fd, 12, SEEK_CUR);
  code = Get_u_char(pk_fd);
  while (code != PK_POST)
    {
      if(code <= PK_FLAG)
	def_char(code);
      else if ((code >= PK_XXX1) && (code <= PK_XXX4))
	{
	  code = Get_u_long_n(pk_fd, (code + 1 - PK_XXX1));
	  lseek(pk_fd, code, SEEK_CUR);
	}
      else if (code == PK_YYY)
	{
	  lseek(pk_fd, 4, SEEK_CUR);
	}
      else if (code == PK_NOP)
	{
	  continue;
	}
      else
	{
	  fprintf(stderr, "Undefined PK command %ld.\n", code);
	  exit(1);
	}
      code = Get_u_char(pk_fd);
    }
  close(pk_fd);
}

static void pk_put_char(void)
{
  long dynf;
  long xoff;
  long yoff;
  long width;
  long width_8;
  long height;
  u_char *font_buffer;

  Enter("pk_put_char");

  if (font->pk->code[c_id].bitmap == NULL)
    {
      dynf = font->pk->code[c_id].dynf;
      
      if (dynf == 14)
	unpack_raster();
      else
	unpack_run();
    }
  
  width = font->pk->code[c_id].width;
  width_8 = font->pk->code[c_id].width_8;
  height = font->pk->code[c_id].height;
  
  if (width == 0)
    return;
  if (height == 0)
    return;
  
#ifndef USE_PTEX
  xoff = (Code.sp_x + Dim.h_sp_offset) / Dim.h_sp_per_dot;
  yoff = (Code.sp_y + Dim.v_sp_offset) / Dim.v_sp_per_dot;
  xoff -= font->pk->code[c_id].xoff;
  yoff -= font->pk->code[c_id].yoff;

  if ((width + xoff) > Buf->width)
    return;
  if ((height + yoff) > Buf->height)
    return;

  font_buffer = (font->pk->code[c_id].bitmap);
  Draw_pattern_a_dot(font_buffer, width_8, height,
		     (xoff / 8), yoff, (xoff % 8));
#else
  if (Code.ptex_dir == 0)
    {
      xoff = (Code.sp_x + Dim.h_sp_offset) / Dim.h_sp_per_dot;
      yoff = (Code.sp_y + Dim.v_sp_offset) / Dim.v_sp_per_dot;
      xoff -= font->pk->code[c_id].xoff;
      yoff -= font->pk->code[c_id].yoff;
      
      if ((width + xoff) > Buf->width)
	return;
      if ((height + yoff) > Buf->height)
	return;

      font_buffer = (font->pk->code[c_id].bitmap);
      Draw_pattern_a_dot(font_buffer, width_8, height,
			 (xoff / 8), yoff, (xoff % 8));
    }
  else
    {
      xoff = (Code.sp_x + Dim.h_sp_offset) / Dim.h_sp_per_dot;
      yoff = (Code.sp_y + Dim.v_sp_offset) / Dim.v_sp_per_dot;
      xoff += font->pk->code[c_id].yoff;
      yoff -= font->pk->code[c_id].xoff;
      
      if ((width + xoff) > Buf->width)
	return;
      if ((height + yoff) > Buf->height)
	return;

      font_buffer = (font->pk->code[c_id].bitmap);
      Draw_rotate_pattern_a_dot(font_buffer, width_8, height,
		      xoff, yoff);
    }
#endif
}

static void def_char(long flagbyte)
{
  long dynf;
  long packetlength;
  long draw_flag;
  long car;
  long xoff;
  long yoff;
  long width;
  long width_8;
  long height;
  long tfm_width;

  Enter("def_char");

  dynf = (flagbyte / 16);
  flagbyte = (flagbyte % 16);
  if (flagbyte >= 8)
    draw_flag = 1;
  else
    draw_flag = 0;
  flagbyte = (flagbyte % 8);
  if(flagbyte == 7)
    {
      packetlength = (Get_long(pk_fd) - 28);
      car = Get_long(pk_fd);
      tfm_width = (long)(ratio * Get_long(pk_fd));
      lseek(pk_fd, 8, SEEK_CUR);
      width = Get_u_long(pk_fd);
      width_8 = ((width + 7) / 8);
      height = Get_u_long(pk_fd);
      xoff = Get_long(pk_fd);
      yoff = Get_long(pk_fd);
    }
  else if(flagbyte > 3)
    {
      packetlength = (((flagbyte - 4) << 16) + Get_u_short(pk_fd) - 13);
      car = Get_u_char(pk_fd);
      tfm_width = (long)(ratio * Get_u_long_n(pk_fd, 3));
      lseek(pk_fd, 2, SEEK_CUR);
      width = Get_u_short(pk_fd);
      width_8 = ((width + 7) / 8);
      height = Get_u_short(pk_fd);
      xoff = Get_short(pk_fd);
      yoff = Get_short(pk_fd);
#ifdef DEBUG
      printf("%s\n", font->name);
      printf("f %3ld pl %3ld car %3ld tfm %3ld ",
	     flagbyte, packetlength, car, tfm_width);
      printf("w %3ld h %3ld xoff %3ld yoff %3ld\n", width, height, xoff, yoff);
#endif
    }
  else
    {
      packetlength = ((flagbyte << 8) + Get_u_char(pk_fd) - 8);
      car = Get_u_char(pk_fd);
      tfm_width = (long)(ratio * Get_u_long_n(pk_fd, 3));
      lseek(pk_fd, 1, SEEK_CUR);
      width = Get_u_char(pk_fd);
      width_8 = ((width + 7) / 8);
      height = Get_u_char(pk_fd);
      xoff = Get_char(pk_fd);
      yoff = Get_char(pk_fd);
    }
#ifdef DEBUG
  fprintf(stderr, "car %ld  pl %ld\n", car, packetlength);
#endif

  font->pk->code[car].dynf = dynf;
  font->pk->code[car].flagbyte = flagbyte;
  font->pk->code[car].draw_flag = draw_flag;
  font->pk->code[car].tfm_width = tfm_width;
  font->pk->code[car].width = width;
  font->pk->code[car].width_8 = width_8;
  font->pk->code[car].height = height;
  font->pk->code[car].xoff = xoff;
  font->pk->code[car].yoff = yoff;
  font->pk->code[car].data = Get_string(pk_fd, packetlength);
}

static void unpack_raster(void)
{
  long x;
  long y;
  long width;
  long width_8;
  long height;
  u_char mask;
  u_char *bitmap;

  Enter("unpack_raster");

  width = font->pk->code[c_id].width;
  width_8 = font->pk->code[c_id].width_8;
  height = font->pk->code[c_id].height;
  
  if (width == 0)
    return;
  if (height == 0)
    return;
  
  if ((bitmap = calloc((width_8 * height), sizeof(u_char))) == NULL)
    {
      fprintf(stderr, "unpack_raster : Can't allocate memory.\n");
      exit(1);
    }
  mask = 0x80;
  for(y=0 ; y<height ; y++)
    {
      for(x=0 ; x<width ; x++)
	{
	  if ((font->pk->code[c_id].data[(y * width + x) / 8]
	       & mask) != 0)
	    *(bitmap + y * width_8 + x / 8) |= (0x80 >> (x % 8));
	  
	  if (mask == 0x01)
	    mask = 0x80;
	  else
	    mask = (mask >> 1);
	}
    }

  font->pk->code[c_id].bitmap = bitmap;
  free(font->pk->code[c_id].data);
  font->pk->code[c_id].data = NULL;
}

static void unpack_run(void)
{
  long dynf;
  long flagbyte;
  long draw_flag;
  long x;
  long x1;
  long y;
  long width;
  long width_8;
  long height;
  long run;
  long repeat;
  long nyb;
  u_char mask;
  u_char *bitmap;

  Enter("unpack_run");

  dynf = font->pk->code[c_id].dynf;
  flagbyte = font->pk->code[c_id].flagbyte;
  draw_flag = font->pk->code[c_id].draw_flag;
  width = font->pk->code[c_id].width;
  width_8 = font->pk->code[c_id].width_8;
  height = font->pk->code[c_id].height;
  
  if (width == 0)
    return;
  if (height == 0)
    return;

  if ((bitmap = calloc((width_8 * height), sizeof(u_char))) == NULL)
    {
      fprintf(stderr, "unpack_run : Can't allocate memory.\n");
      exit(1);
    }

  run = 0;
  nyb = 0;
  nyb_count = 0;
  repeat = 0;

  if (draw_flag == 0)
    draw_flag++;
  else
    draw_flag--;

  y = 0;
  while (y < height)
    {
      mask = 0x80;
      x = 0;
      repeat = 0;
      while (x < width)
	{
#if 1
	  if (run == 0)
	    {
	      if (draw_flag == 0)
		draw_flag++;
	      else
		draw_flag--;

	      nyb = get_nyb();
	      if (nyb > 14)
		{
		  repeat = ((nyb - 14) * 2 - 1);
		  nyb = get_nyb();
		}
	      else if (nyb == 14)
		{
		  nyb = get_nyb();
		  if (nyb > dynf)
		    {
		      repeat = (nyb - dynf - 1) * 16;
		      repeat += (get_nyb() + dynf + 1);
		    }
		  else if (nyb > 0)
		    {
		      repeat = nyb;
		    }
		  else
		    {
		      do
			{
			  nyb++;
			  repeat = get_nyb();
			} while(repeat == 0);
		      while (nyb > 0)
			{
			  repeat = repeat * 16 + get_nyb();
			  nyb--;
			}
		      repeat += (13 - dynf) * 16 + dynf - 15;
		    }
		  nyb = get_nyb();
		}

	      if (nyb > dynf)
		{
		  run = (nyb - dynf - 1) * 16;
		  run += (get_nyb() + dynf + 1);
		}
	      else if (nyb > 0)
		{
		  run = nyb;
		}
	      else
		{
		  do
		    {
		      nyb++;
		      run = get_nyb();
		    } while (run == 0);
		  while (nyb > 0)
		    {
		      run = run * 16 + get_nyb();
		      nyb--;
		    }
		  run += (13 - dynf) * 16 + dynf - 15;
		}
	    }
#else
	  if (run == 0)
	    {
	      if (draw_flag == 0)
		draw_flag++;
	      else
		draw_flag--;

	      for ( ; ; )
		{
		  nyb = get_nyb();
		  if (nyb >= 14)
		    {
		      repeat = (nyb * 2 - 29);
		      continue;
		    }
		  else if (nyb > dynf)
		    {
		      run = (nyb - dynf - 1) * 16;
		      run += (get_nyb() + dynf + 1);
		    }
		  else if (nyb > 0)
		    {
		      run = nyb;
		    }
		  else /* nyb == 0 */
		    {
		      do
			{
			  nyb++;
			  run = get_nyb();
			} while(run == 0);

		      while (nyb > 0)
			{
			  run = run * 16 + get_nyb();
			  nyb--;
			}
		      run += (13 - dynf) * 16 + dynf - 15;
		    }

		  if (repeat >= 0)
		    break;

		  repeat = run;
		}
	    }
#endif
	  if ((x + run) <= width)
	    x1 = (x + run);
	  else
	    x1 = width;
	  run -= (x1 - x);

	  while (x < x1)
	    {
	      if (draw_flag == 1)
		*(bitmap + y * width_8 + (x / 8)) |= mask;
	      if (mask == 0x01)
		mask = 0x80;
	      else
		mask = (mask >> 1);
	      x++;
	    }
	}

      while (repeat > 0)
	{
	  y++;
	  for (x=0 ; x<width_8 ; x++)
	    *(bitmap + y * width_8 + x) = *(bitmap + (y - 1) * width_8 + x);
	  repeat--;
	}
      y++;
    }

  font->pk->code[c_id].bitmap = bitmap;
}

static long get_nyb(void)
{
  static long nyb;

  Enter("get_nyb");

  if ((nyb_count % 2) == 0)
    nyb = (long)(font->pk->code[c_id].data[nyb_count / 2] >> 4);
  else
    nyb = (long)(font->pk->code[c_id].data[nyb_count / 2] & 0x0f);
  nyb_count++;

  return nyb;
}

/* End of font_pk.c */
